# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: signing.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor.FileDescriptor(
    name="signing.proto",
    package="fetch.aea.Signing",
    syntax="proto3",
    serialized_options=None,
    serialized_pb=b"\n\rsigning.proto\x12\x11\x66\x65tch.aea.Signing\"\xa8\x0c\n\x0eSigningMessage\x12\x12\n\nmessage_id\x18\x01 \x01(\x05\x12\"\n\x1a\x64ialogue_starter_reference\x18\x02 \x01(\t\x12$\n\x1c\x64ialogue_responder_reference\x18\x03 \x01(\t\x12\x0e\n\x06target\x18\x04 \x01(\x05\x12\x45\n\x05\x65rror\x18\x05 \x01(\x0b\x32\x34.fetch.aea.Signing.SigningMessage.Error_PerformativeH\x00\x12S\n\x0csign_message\x18\x06 \x01(\x0b\x32;.fetch.aea.Signing.SigningMessage.Sign_Message_PerformativeH\x00\x12[\n\x10sign_transaction\x18\x07 \x01(\x0b\x32?.fetch.aea.Signing.SigningMessage.Sign_Transaction_PerformativeH\x00\x12W\n\x0esigned_message\x18\x08 \x01(\x0b\x32=.fetch.aea.Signing.SigningMessage.Signed_Message_PerformativeH\x00\x12_\n\x12signed_transaction\x18\t \x01(\x0b\x32\x41.fetch.aea.Signing.SigningMessage.Signed_Transaction_PerformativeH\x00\x1a\xb3\x01\n\tErrorCode\x12M\n\nerror_code\x18\x01 \x01(\x0e\x32\x39.fetch.aea.Signing.SigningMessage.ErrorCode.ErrorCodeEnum\"W\n\rErrorCodeEnum\x12 \n\x1cUNSUCCESSFUL_MESSAGE_SIGNING\x10\x00\x12$\n UNSUCCESSFUL_TRANSACTION_SIGNING\x10\x01\x1a!\n\nRawMessage\x12\x13\n\x0braw_message\x18\x01 \x01(\x0c\x1a)\n\x0eRawTransaction\x12\x17\n\x0fraw_transaction\x18\x01 \x01(\x0c\x1a'\n\rSignedMessage\x12\x16\n\x0esigned_message\x18\x01 \x01(\x0c\x1a/\n\x11SignedTransaction\x12\x1a\n\x12signed_transaction\x18\x01 \x01(\x0c\x1a\x16\n\x05Terms\x12\r\n\x05terms\x18\x01 \x01(\x0c\x1a\xa2\x01\n\x1dSign_Transaction_Performative\x12\x36\n\x05terms\x18\x01 \x01(\x0b\x32'.fetch.aea.Signing.SigningMessage.Terms\x12I\n\x0fraw_transaction\x18\x02 \x01(\x0b\x32\x30.fetch.aea.Signing.SigningMessage.RawTransaction\x1a\x96\x01\n\x19Sign_Message_Performative\x12\x36\n\x05terms\x18\x01 \x01(\x0b\x32'.fetch.aea.Signing.SigningMessage.Terms\x12\x41\n\x0braw_message\x18\x02 \x01(\x0b\x32,.fetch.aea.Signing.SigningMessage.RawMessage\x1ar\n\x1fSigned_Transaction_Performative\x12O\n\x12signed_transaction\x18\x01 \x01(\x0b\x32\x33.fetch.aea.Signing.SigningMessage.SignedTransaction\x1a\x66\n\x1bSigned_Message_Performative\x12G\n\x0esigned_message\x18\x01 \x01(\x0b\x32/.fetch.aea.Signing.SigningMessage.SignedMessage\x1aU\n\x12\x45rror_Performative\x12?\n\nerror_code\x18\x01 \x01(\x0b\x32+.fetch.aea.Signing.SigningMessage.ErrorCodeB\x0e\n\x0cperformativeb\x06proto3",
)


_SIGNINGMESSAGE_ERRORCODE_ERRORCODEENUM = _descriptor.EnumDescriptor(
    name="ErrorCodeEnum",
    full_name="fetch.aea.Signing.SigningMessage.ErrorCode.ErrorCodeEnum",
    filename=None,
    file=DESCRIPTOR,
    values=[
        _descriptor.EnumValueDescriptor(
            name="UNSUCCESSFUL_MESSAGE_SIGNING",
            index=0,
            number=0,
            serialized_options=None,
            type=None,
        ),
        _descriptor.EnumValueDescriptor(
            name="UNSUCCESSFUL_TRANSACTION_SIGNING",
            index=1,
            number=1,
            serialized_options=None,
            type=None,
        ),
    ],
    containing_type=None,
    serialized_options=None,
    serialized_start=693,
    serialized_end=780,
)
_sym_db.RegisterEnumDescriptor(_SIGNINGMESSAGE_ERRORCODE_ERRORCODEENUM)


_SIGNINGMESSAGE_ERRORCODE = _descriptor.Descriptor(
    name="ErrorCode",
    full_name="fetch.aea.Signing.SigningMessage.ErrorCode",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="error_code",
            full_name="fetch.aea.Signing.SigningMessage.ErrorCode.error_code",
            index=0,
            number=1,
            type=14,
            cpp_type=8,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[_SIGNINGMESSAGE_ERRORCODE_ERRORCODEENUM,],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=601,
    serialized_end=780,
)

_SIGNINGMESSAGE_RAWMESSAGE = _descriptor.Descriptor(
    name="RawMessage",
    full_name="fetch.aea.Signing.SigningMessage.RawMessage",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="raw_message",
            full_name="fetch.aea.Signing.SigningMessage.RawMessage.raw_message",
            index=0,
            number=1,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=782,
    serialized_end=815,
)

_SIGNINGMESSAGE_RAWTRANSACTION = _descriptor.Descriptor(
    name="RawTransaction",
    full_name="fetch.aea.Signing.SigningMessage.RawTransaction",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="raw_transaction",
            full_name="fetch.aea.Signing.SigningMessage.RawTransaction.raw_transaction",
            index=0,
            number=1,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=817,
    serialized_end=858,
)

_SIGNINGMESSAGE_SIGNEDMESSAGE = _descriptor.Descriptor(
    name="SignedMessage",
    full_name="fetch.aea.Signing.SigningMessage.SignedMessage",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="signed_message",
            full_name="fetch.aea.Signing.SigningMessage.SignedMessage.signed_message",
            index=0,
            number=1,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=860,
    serialized_end=899,
)

_SIGNINGMESSAGE_SIGNEDTRANSACTION = _descriptor.Descriptor(
    name="SignedTransaction",
    full_name="fetch.aea.Signing.SigningMessage.SignedTransaction",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="signed_transaction",
            full_name="fetch.aea.Signing.SigningMessage.SignedTransaction.signed_transaction",
            index=0,
            number=1,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=901,
    serialized_end=948,
)

_SIGNINGMESSAGE_TERMS = _descriptor.Descriptor(
    name="Terms",
    full_name="fetch.aea.Signing.SigningMessage.Terms",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="terms",
            full_name="fetch.aea.Signing.SigningMessage.Terms.terms",
            index=0,
            number=1,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=950,
    serialized_end=972,
)

_SIGNINGMESSAGE_SIGN_TRANSACTION_PERFORMATIVE = _descriptor.Descriptor(
    name="Sign_Transaction_Performative",
    full_name="fetch.aea.Signing.SigningMessage.Sign_Transaction_Performative",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="terms",
            full_name="fetch.aea.Signing.SigningMessage.Sign_Transaction_Performative.terms",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="raw_transaction",
            full_name="fetch.aea.Signing.SigningMessage.Sign_Transaction_Performative.raw_transaction",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=975,
    serialized_end=1137,
)

_SIGNINGMESSAGE_SIGN_MESSAGE_PERFORMATIVE = _descriptor.Descriptor(
    name="Sign_Message_Performative",
    full_name="fetch.aea.Signing.SigningMessage.Sign_Message_Performative",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="terms",
            full_name="fetch.aea.Signing.SigningMessage.Sign_Message_Performative.terms",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="raw_message",
            full_name="fetch.aea.Signing.SigningMessage.Sign_Message_Performative.raw_message",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1140,
    serialized_end=1290,
)

_SIGNINGMESSAGE_SIGNED_TRANSACTION_PERFORMATIVE = _descriptor.Descriptor(
    name="Signed_Transaction_Performative",
    full_name="fetch.aea.Signing.SigningMessage.Signed_Transaction_Performative",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="signed_transaction",
            full_name="fetch.aea.Signing.SigningMessage.Signed_Transaction_Performative.signed_transaction",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1292,
    serialized_end=1406,
)

_SIGNINGMESSAGE_SIGNED_MESSAGE_PERFORMATIVE = _descriptor.Descriptor(
    name="Signed_Message_Performative",
    full_name="fetch.aea.Signing.SigningMessage.Signed_Message_Performative",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="signed_message",
            full_name="fetch.aea.Signing.SigningMessage.Signed_Message_Performative.signed_message",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1408,
    serialized_end=1510,
)

_SIGNINGMESSAGE_ERROR_PERFORMATIVE = _descriptor.Descriptor(
    name="Error_Performative",
    full_name="fetch.aea.Signing.SigningMessage.Error_Performative",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="error_code",
            full_name="fetch.aea.Signing.SigningMessage.Error_Performative.error_code",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1512,
    serialized_end=1597,
)

_SIGNINGMESSAGE = _descriptor.Descriptor(
    name="SigningMessage",
    full_name="fetch.aea.Signing.SigningMessage",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="message_id",
            full_name="fetch.aea.Signing.SigningMessage.message_id",
            index=0,
            number=1,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="dialogue_starter_reference",
            full_name="fetch.aea.Signing.SigningMessage.dialogue_starter_reference",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="dialogue_responder_reference",
            full_name="fetch.aea.Signing.SigningMessage.dialogue_responder_reference",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="target",
            full_name="fetch.aea.Signing.SigningMessage.target",
            index=3,
            number=4,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="error",
            full_name="fetch.aea.Signing.SigningMessage.error",
            index=4,
            number=5,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="sign_message",
            full_name="fetch.aea.Signing.SigningMessage.sign_message",
            index=5,
            number=6,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="sign_transaction",
            full_name="fetch.aea.Signing.SigningMessage.sign_transaction",
            index=6,
            number=7,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="signed_message",
            full_name="fetch.aea.Signing.SigningMessage.signed_message",
            index=7,
            number=8,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="signed_transaction",
            full_name="fetch.aea.Signing.SigningMessage.signed_transaction",
            index=8,
            number=9,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[
        _SIGNINGMESSAGE_ERRORCODE,
        _SIGNINGMESSAGE_RAWMESSAGE,
        _SIGNINGMESSAGE_RAWTRANSACTION,
        _SIGNINGMESSAGE_SIGNEDMESSAGE,
        _SIGNINGMESSAGE_SIGNEDTRANSACTION,
        _SIGNINGMESSAGE_TERMS,
        _SIGNINGMESSAGE_SIGN_TRANSACTION_PERFORMATIVE,
        _SIGNINGMESSAGE_SIGN_MESSAGE_PERFORMATIVE,
        _SIGNINGMESSAGE_SIGNED_TRANSACTION_PERFORMATIVE,
        _SIGNINGMESSAGE_SIGNED_MESSAGE_PERFORMATIVE,
        _SIGNINGMESSAGE_ERROR_PERFORMATIVE,
    ],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[
        _descriptor.OneofDescriptor(
            name="performative",
            full_name="fetch.aea.Signing.SigningMessage.performative",
            index=0,
            containing_type=None,
            fields=[],
        ),
    ],
    serialized_start=37,
    serialized_end=1613,
)

_SIGNINGMESSAGE_ERRORCODE.fields_by_name[
    "error_code"
].enum_type = _SIGNINGMESSAGE_ERRORCODE_ERRORCODEENUM
_SIGNINGMESSAGE_ERRORCODE.containing_type = _SIGNINGMESSAGE
_SIGNINGMESSAGE_ERRORCODE_ERRORCODEENUM.containing_type = _SIGNINGMESSAGE_ERRORCODE
_SIGNINGMESSAGE_RAWMESSAGE.containing_type = _SIGNINGMESSAGE
_SIGNINGMESSAGE_RAWTRANSACTION.containing_type = _SIGNINGMESSAGE
_SIGNINGMESSAGE_SIGNEDMESSAGE.containing_type = _SIGNINGMESSAGE
_SIGNINGMESSAGE_SIGNEDTRANSACTION.containing_type = _SIGNINGMESSAGE
_SIGNINGMESSAGE_TERMS.containing_type = _SIGNINGMESSAGE
_SIGNINGMESSAGE_SIGN_TRANSACTION_PERFORMATIVE.fields_by_name[
    "terms"
].message_type = _SIGNINGMESSAGE_TERMS
_SIGNINGMESSAGE_SIGN_TRANSACTION_PERFORMATIVE.fields_by_name[
    "raw_transaction"
].message_type = _SIGNINGMESSAGE_RAWTRANSACTION
_SIGNINGMESSAGE_SIGN_TRANSACTION_PERFORMATIVE.containing_type = _SIGNINGMESSAGE
_SIGNINGMESSAGE_SIGN_MESSAGE_PERFORMATIVE.fields_by_name[
    "terms"
].message_type = _SIGNINGMESSAGE_TERMS
_SIGNINGMESSAGE_SIGN_MESSAGE_PERFORMATIVE.fields_by_name[
    "raw_message"
].message_type = _SIGNINGMESSAGE_RAWMESSAGE
_SIGNINGMESSAGE_SIGN_MESSAGE_PERFORMATIVE.containing_type = _SIGNINGMESSAGE
_SIGNINGMESSAGE_SIGNED_TRANSACTION_PERFORMATIVE.fields_by_name[
    "signed_transaction"
].message_type = _SIGNINGMESSAGE_SIGNEDTRANSACTION
_SIGNINGMESSAGE_SIGNED_TRANSACTION_PERFORMATIVE.containing_type = _SIGNINGMESSAGE
_SIGNINGMESSAGE_SIGNED_MESSAGE_PERFORMATIVE.fields_by_name[
    "signed_message"
].message_type = _SIGNINGMESSAGE_SIGNEDMESSAGE
_SIGNINGMESSAGE_SIGNED_MESSAGE_PERFORMATIVE.containing_type = _SIGNINGMESSAGE
_SIGNINGMESSAGE_ERROR_PERFORMATIVE.fields_by_name[
    "error_code"
].message_type = _SIGNINGMESSAGE_ERRORCODE
_SIGNINGMESSAGE_ERROR_PERFORMATIVE.containing_type = _SIGNINGMESSAGE
_SIGNINGMESSAGE.fields_by_name[
    "error"
].message_type = _SIGNINGMESSAGE_ERROR_PERFORMATIVE
_SIGNINGMESSAGE.fields_by_name[
    "sign_message"
].message_type = _SIGNINGMESSAGE_SIGN_MESSAGE_PERFORMATIVE
_SIGNINGMESSAGE.fields_by_name[
    "sign_transaction"
].message_type = _SIGNINGMESSAGE_SIGN_TRANSACTION_PERFORMATIVE
_SIGNINGMESSAGE.fields_by_name[
    "signed_message"
].message_type = _SIGNINGMESSAGE_SIGNED_MESSAGE_PERFORMATIVE
_SIGNINGMESSAGE.fields_by_name[
    "signed_transaction"
].message_type = _SIGNINGMESSAGE_SIGNED_TRANSACTION_PERFORMATIVE
_SIGNINGMESSAGE.oneofs_by_name["performative"].fields.append(
    _SIGNINGMESSAGE.fields_by_name["error"]
)
_SIGNINGMESSAGE.fields_by_name[
    "error"
].containing_oneof = _SIGNINGMESSAGE.oneofs_by_name["performative"]
_SIGNINGMESSAGE.oneofs_by_name["performative"].fields.append(
    _SIGNINGMESSAGE.fields_by_name["sign_message"]
)
_SIGNINGMESSAGE.fields_by_name[
    "sign_message"
].containing_oneof = _SIGNINGMESSAGE.oneofs_by_name["performative"]
_SIGNINGMESSAGE.oneofs_by_name["performative"].fields.append(
    _SIGNINGMESSAGE.fields_by_name["sign_transaction"]
)
_SIGNINGMESSAGE.fields_by_name[
    "sign_transaction"
].containing_oneof = _SIGNINGMESSAGE.oneofs_by_name["performative"]
_SIGNINGMESSAGE.oneofs_by_name["performative"].fields.append(
    _SIGNINGMESSAGE.fields_by_name["signed_message"]
)
_SIGNINGMESSAGE.fields_by_name[
    "signed_message"
].containing_oneof = _SIGNINGMESSAGE.oneofs_by_name["performative"]
_SIGNINGMESSAGE.oneofs_by_name["performative"].fields.append(
    _SIGNINGMESSAGE.fields_by_name["signed_transaction"]
)
_SIGNINGMESSAGE.fields_by_name[
    "signed_transaction"
].containing_oneof = _SIGNINGMESSAGE.oneofs_by_name["performative"]
DESCRIPTOR.message_types_by_name["SigningMessage"] = _SIGNINGMESSAGE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

SigningMessage = _reflection.GeneratedProtocolMessageType(
    "SigningMessage",
    (_message.Message,),
    {
        "ErrorCode": _reflection.GeneratedProtocolMessageType(
            "ErrorCode",
            (_message.Message,),
            {
                "DESCRIPTOR": _SIGNINGMESSAGE_ERRORCODE,
                "__module__": "signing_pb2"
                # @@protoc_insertion_point(class_scope:fetch.aea.Signing.SigningMessage.ErrorCode)
            },
        ),
        "RawMessage": _reflection.GeneratedProtocolMessageType(
            "RawMessage",
            (_message.Message,),
            {
                "DESCRIPTOR": _SIGNINGMESSAGE_RAWMESSAGE,
                "__module__": "signing_pb2"
                # @@protoc_insertion_point(class_scope:fetch.aea.Signing.SigningMessage.RawMessage)
            },
        ),
        "RawTransaction": _reflection.GeneratedProtocolMessageType(
            "RawTransaction",
            (_message.Message,),
            {
                "DESCRIPTOR": _SIGNINGMESSAGE_RAWTRANSACTION,
                "__module__": "signing_pb2"
                # @@protoc_insertion_point(class_scope:fetch.aea.Signing.SigningMessage.RawTransaction)
            },
        ),
        "SignedMessage": _reflection.GeneratedProtocolMessageType(
            "SignedMessage",
            (_message.Message,),
            {
                "DESCRIPTOR": _SIGNINGMESSAGE_SIGNEDMESSAGE,
                "__module__": "signing_pb2"
                # @@protoc_insertion_point(class_scope:fetch.aea.Signing.SigningMessage.SignedMessage)
            },
        ),
        "SignedTransaction": _reflection.GeneratedProtocolMessageType(
            "SignedTransaction",
            (_message.Message,),
            {
                "DESCRIPTOR": _SIGNINGMESSAGE_SIGNEDTRANSACTION,
                "__module__": "signing_pb2"
                # @@protoc_insertion_point(class_scope:fetch.aea.Signing.SigningMessage.SignedTransaction)
            },
        ),
        "Terms": _reflection.GeneratedProtocolMessageType(
            "Terms",
            (_message.Message,),
            {
                "DESCRIPTOR": _SIGNINGMESSAGE_TERMS,
                "__module__": "signing_pb2"
                # @@protoc_insertion_point(class_scope:fetch.aea.Signing.SigningMessage.Terms)
            },
        ),
        "Sign_Transaction_Performative": _reflection.GeneratedProtocolMessageType(
            "Sign_Transaction_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _SIGNINGMESSAGE_SIGN_TRANSACTION_PERFORMATIVE,
                "__module__": "signing_pb2"
                # @@protoc_insertion_point(class_scope:fetch.aea.Signing.SigningMessage.Sign_Transaction_Performative)
            },
        ),
        "Sign_Message_Performative": _reflection.GeneratedProtocolMessageType(
            "Sign_Message_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _SIGNINGMESSAGE_SIGN_MESSAGE_PERFORMATIVE,
                "__module__": "signing_pb2"
                # @@protoc_insertion_point(class_scope:fetch.aea.Signing.SigningMessage.Sign_Message_Performative)
            },
        ),
        "Signed_Transaction_Performative": _reflection.GeneratedProtocolMessageType(
            "Signed_Transaction_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _SIGNINGMESSAGE_SIGNED_TRANSACTION_PERFORMATIVE,
                "__module__": "signing_pb2"
                # @@protoc_insertion_point(class_scope:fetch.aea.Signing.SigningMessage.Signed_Transaction_Performative)
            },
        ),
        "Signed_Message_Performative": _reflection.GeneratedProtocolMessageType(
            "Signed_Message_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _SIGNINGMESSAGE_SIGNED_MESSAGE_PERFORMATIVE,
                "__module__": "signing_pb2"
                # @@protoc_insertion_point(class_scope:fetch.aea.Signing.SigningMessage.Signed_Message_Performative)
            },
        ),
        "Error_Performative": _reflection.GeneratedProtocolMessageType(
            "Error_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _SIGNINGMESSAGE_ERROR_PERFORMATIVE,
                "__module__": "signing_pb2"
                # @@protoc_insertion_point(class_scope:fetch.aea.Signing.SigningMessage.Error_Performative)
            },
        ),
        "DESCRIPTOR": _SIGNINGMESSAGE,
        "__module__": "signing_pb2"
        # @@protoc_insertion_point(class_scope:fetch.aea.Signing.SigningMessage)
    },
)
_sym_db.RegisterMessage(SigningMessage)
_sym_db.RegisterMessage(SigningMessage.ErrorCode)
_sym_db.RegisterMessage(SigningMessage.RawMessage)
_sym_db.RegisterMessage(SigningMessage.RawTransaction)
_sym_db.RegisterMessage(SigningMessage.SignedMessage)
_sym_db.RegisterMessage(SigningMessage.SignedTransaction)
_sym_db.RegisterMessage(SigningMessage.Terms)
_sym_db.RegisterMessage(SigningMessage.Sign_Transaction_Performative)
_sym_db.RegisterMessage(SigningMessage.Sign_Message_Performative)
_sym_db.RegisterMessage(SigningMessage.Signed_Transaction_Performative)
_sym_db.RegisterMessage(SigningMessage.Signed_Message_Performative)
_sym_db.RegisterMessage(SigningMessage.Error_Performative)


# @@protoc_insertion_point(module_scope)
