# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ledger_api.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor.FileDescriptor(
    name="ledger_api.proto",
    package="fetch.aea.LedgerApi",
    syntax="proto3",
    serialized_options=None,
    serialized_pb=b'\n\x10ledger_api.proto\x12\x13\x66\x65tch.aea.LedgerApi"\xf1\x10\n\x10LedgerApiMessage\x12\x12\n\nmessage_id\x18\x01 \x01(\x05\x12"\n\x1a\x64ialogue_starter_reference\x18\x02 \x01(\t\x12$\n\x1c\x64ialogue_responder_reference\x18\x03 \x01(\t\x12\x0e\n\x06target\x18\x04 \x01(\x05\x12M\n\x07\x62\x61lance\x18\x05 \x01(\x0b\x32:.fetch.aea.LedgerApi.LedgerApiMessage.Balance_PerformativeH\x00\x12I\n\x05\x65rror\x18\x06 \x01(\x0b\x32\x38.fetch.aea.LedgerApi.LedgerApiMessage.Error_PerformativeH\x00\x12U\n\x0bget_balance\x18\x07 \x01(\x0b\x32>.fetch.aea.LedgerApi.LedgerApiMessage.Get_Balance_PerformativeH\x00\x12\x65\n\x13get_raw_transaction\x18\x08 \x01(\x0b\x32\x46.fetch.aea.LedgerApi.LedgerApiMessage.Get_Raw_Transaction_PerformativeH\x00\x12m\n\x17get_transaction_receipt\x18\t \x01(\x0b\x32J.fetch.aea.LedgerApi.LedgerApiMessage.Get_Transaction_Receipt_PerformativeH\x00\x12]\n\x0fraw_transaction\x18\n \x01(\x0b\x32\x42.fetch.aea.LedgerApi.LedgerApiMessage.Raw_Transaction_PerformativeH\x00\x12m\n\x17send_signed_transaction\x18\x0b \x01(\x0b\x32J.fetch.aea.LedgerApi.LedgerApiMessage.Send_Signed_Transaction_PerformativeH\x00\x12\x63\n\x12transaction_digest\x18\x0c \x01(\x0b\x32\x45.fetch.aea.LedgerApi.LedgerApiMessage.Transaction_Digest_PerformativeH\x00\x12\x65\n\x13transaction_receipt\x18\r \x01(\x0b\x32\x46.fetch.aea.LedgerApi.LedgerApiMessage.Transaction_Receipt_PerformativeH\x00\x1a)\n\x0eRawTransaction\x12\x17\n\x0fraw_transaction\x18\x01 \x01(\x0c\x1a/\n\x11SignedTransaction\x12\x1a\n\x12signed_transaction\x18\x01 \x01(\x0c\x1a\x16\n\x05Terms\x12\r\n\x05terms\x18\x01 \x01(\x0c\x1a/\n\x11TransactionDigest\x12\x1a\n\x12transaction_digest\x18\x01 \x01(\x0c\x1a\x31\n\x12TransactionReceipt\x12\x1b\n\x13transaction_receipt\x18\x01 \x01(\x0c\x1a>\n\x18Get_Balance_Performative\x12\x11\n\tledger_id\x18\x01 \x01(\t\x12\x0f\n\x07\x61\x64\x64ress\x18\x02 \x01(\t\x1a^\n Get_Raw_Transaction_Performative\x12:\n\x05terms\x18\x01 \x01(\x0b\x32+.fetch.aea.LedgerApi.LedgerApiMessage.Terms\x1a{\n$Send_Signed_Transaction_Performative\x12S\n\x12signed_transaction\x18\x01 \x01(\x0b\x32\x37.fetch.aea.LedgerApi.LedgerApiMessage.SignedTransaction\x1a{\n$Get_Transaction_Receipt_Performative\x12S\n\x12transaction_digest\x18\x01 \x01(\x0b\x32\x37.fetch.aea.LedgerApi.LedgerApiMessage.TransactionDigest\x1a:\n\x14\x42\x61lance_Performative\x12\x11\n\tledger_id\x18\x01 \x01(\t\x12\x0f\n\x07\x62\x61lance\x18\x02 \x01(\x05\x1am\n\x1cRaw_Transaction_Performative\x12M\n\x0fraw_transaction\x18\x01 \x01(\x0b\x32\x34.fetch.aea.LedgerApi.LedgerApiMessage.RawTransaction\x1av\n\x1fTransaction_Digest_Performative\x12S\n\x12transaction_digest\x18\x01 \x01(\x0b\x32\x37.fetch.aea.LedgerApi.LedgerApiMessage.TransactionDigest\x1ay\n Transaction_Receipt_Performative\x12U\n\x13transaction_receipt\x18\x01 \x01(\x0b\x32\x38.fetch.aea.LedgerApi.LedgerApiMessage.TransactionReceipt\x1an\n\x12\x45rror_Performative\x12\x0c\n\x04\x63ode\x18\x01 \x01(\x05\x12\x0f\n\x07message\x18\x02 \x01(\t\x12\x16\n\x0emessage_is_set\x18\x03 \x01(\x08\x12\x0c\n\x04\x64\x61ta\x18\x04 \x01(\x0c\x12\x13\n\x0b\x64\x61ta_is_set\x18\x05 \x01(\x08\x42\x0e\n\x0cperformativeb\x06proto3',
)


_LEDGERAPIMESSAGE_RAWTRANSACTION = _descriptor.Descriptor(
    name="RawTransaction",
    full_name="fetch.aea.LedgerApi.LedgerApiMessage.RawTransaction",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="raw_transaction",
            full_name="fetch.aea.LedgerApi.LedgerApiMessage.RawTransaction.raw_transaction",
            index=0,
            number=1,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1037,
    serialized_end=1078,
)

_LEDGERAPIMESSAGE_SIGNEDTRANSACTION = _descriptor.Descriptor(
    name="SignedTransaction",
    full_name="fetch.aea.LedgerApi.LedgerApiMessage.SignedTransaction",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="signed_transaction",
            full_name="fetch.aea.LedgerApi.LedgerApiMessage.SignedTransaction.signed_transaction",
            index=0,
            number=1,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1080,
    serialized_end=1127,
)

_LEDGERAPIMESSAGE_TERMS = _descriptor.Descriptor(
    name="Terms",
    full_name="fetch.aea.LedgerApi.LedgerApiMessage.Terms",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="terms",
            full_name="fetch.aea.LedgerApi.LedgerApiMessage.Terms.terms",
            index=0,
            number=1,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1129,
    serialized_end=1151,
)

_LEDGERAPIMESSAGE_TRANSACTIONDIGEST = _descriptor.Descriptor(
    name="TransactionDigest",
    full_name="fetch.aea.LedgerApi.LedgerApiMessage.TransactionDigest",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="transaction_digest",
            full_name="fetch.aea.LedgerApi.LedgerApiMessage.TransactionDigest.transaction_digest",
            index=0,
            number=1,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1153,
    serialized_end=1200,
)

_LEDGERAPIMESSAGE_TRANSACTIONRECEIPT = _descriptor.Descriptor(
    name="TransactionReceipt",
    full_name="fetch.aea.LedgerApi.LedgerApiMessage.TransactionReceipt",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="transaction_receipt",
            full_name="fetch.aea.LedgerApi.LedgerApiMessage.TransactionReceipt.transaction_receipt",
            index=0,
            number=1,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1202,
    serialized_end=1251,
)

_LEDGERAPIMESSAGE_GET_BALANCE_PERFORMATIVE = _descriptor.Descriptor(
    name="Get_Balance_Performative",
    full_name="fetch.aea.LedgerApi.LedgerApiMessage.Get_Balance_Performative",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="ledger_id",
            full_name="fetch.aea.LedgerApi.LedgerApiMessage.Get_Balance_Performative.ledger_id",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="address",
            full_name="fetch.aea.LedgerApi.LedgerApiMessage.Get_Balance_Performative.address",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1253,
    serialized_end=1315,
)

_LEDGERAPIMESSAGE_GET_RAW_TRANSACTION_PERFORMATIVE = _descriptor.Descriptor(
    name="Get_Raw_Transaction_Performative",
    full_name="fetch.aea.LedgerApi.LedgerApiMessage.Get_Raw_Transaction_Performative",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="terms",
            full_name="fetch.aea.LedgerApi.LedgerApiMessage.Get_Raw_Transaction_Performative.terms",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1317,
    serialized_end=1411,
)

_LEDGERAPIMESSAGE_SEND_SIGNED_TRANSACTION_PERFORMATIVE = _descriptor.Descriptor(
    name="Send_Signed_Transaction_Performative",
    full_name="fetch.aea.LedgerApi.LedgerApiMessage.Send_Signed_Transaction_Performative",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="signed_transaction",
            full_name="fetch.aea.LedgerApi.LedgerApiMessage.Send_Signed_Transaction_Performative.signed_transaction",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1413,
    serialized_end=1536,
)

_LEDGERAPIMESSAGE_GET_TRANSACTION_RECEIPT_PERFORMATIVE = _descriptor.Descriptor(
    name="Get_Transaction_Receipt_Performative",
    full_name="fetch.aea.LedgerApi.LedgerApiMessage.Get_Transaction_Receipt_Performative",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="transaction_digest",
            full_name="fetch.aea.LedgerApi.LedgerApiMessage.Get_Transaction_Receipt_Performative.transaction_digest",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1538,
    serialized_end=1661,
)

_LEDGERAPIMESSAGE_BALANCE_PERFORMATIVE = _descriptor.Descriptor(
    name="Balance_Performative",
    full_name="fetch.aea.LedgerApi.LedgerApiMessage.Balance_Performative",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="ledger_id",
            full_name="fetch.aea.LedgerApi.LedgerApiMessage.Balance_Performative.ledger_id",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="balance",
            full_name="fetch.aea.LedgerApi.LedgerApiMessage.Balance_Performative.balance",
            index=1,
            number=2,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1663,
    serialized_end=1721,
)

_LEDGERAPIMESSAGE_RAW_TRANSACTION_PERFORMATIVE = _descriptor.Descriptor(
    name="Raw_Transaction_Performative",
    full_name="fetch.aea.LedgerApi.LedgerApiMessage.Raw_Transaction_Performative",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="raw_transaction",
            full_name="fetch.aea.LedgerApi.LedgerApiMessage.Raw_Transaction_Performative.raw_transaction",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1723,
    serialized_end=1832,
)

_LEDGERAPIMESSAGE_TRANSACTION_DIGEST_PERFORMATIVE = _descriptor.Descriptor(
    name="Transaction_Digest_Performative",
    full_name="fetch.aea.LedgerApi.LedgerApiMessage.Transaction_Digest_Performative",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="transaction_digest",
            full_name="fetch.aea.LedgerApi.LedgerApiMessage.Transaction_Digest_Performative.transaction_digest",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1834,
    serialized_end=1952,
)

_LEDGERAPIMESSAGE_TRANSACTION_RECEIPT_PERFORMATIVE = _descriptor.Descriptor(
    name="Transaction_Receipt_Performative",
    full_name="fetch.aea.LedgerApi.LedgerApiMessage.Transaction_Receipt_Performative",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="transaction_receipt",
            full_name="fetch.aea.LedgerApi.LedgerApiMessage.Transaction_Receipt_Performative.transaction_receipt",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1954,
    serialized_end=2075,
)

_LEDGERAPIMESSAGE_ERROR_PERFORMATIVE = _descriptor.Descriptor(
    name="Error_Performative",
    full_name="fetch.aea.LedgerApi.LedgerApiMessage.Error_Performative",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="code",
            full_name="fetch.aea.LedgerApi.LedgerApiMessage.Error_Performative.code",
            index=0,
            number=1,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="message",
            full_name="fetch.aea.LedgerApi.LedgerApiMessage.Error_Performative.message",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="message_is_set",
            full_name="fetch.aea.LedgerApi.LedgerApiMessage.Error_Performative.message_is_set",
            index=2,
            number=3,
            type=8,
            cpp_type=7,
            label=1,
            has_default_value=False,
            default_value=False,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="data",
            full_name="fetch.aea.LedgerApi.LedgerApiMessage.Error_Performative.data",
            index=3,
            number=4,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="data_is_set",
            full_name="fetch.aea.LedgerApi.LedgerApiMessage.Error_Performative.data_is_set",
            index=4,
            number=5,
            type=8,
            cpp_type=7,
            label=1,
            has_default_value=False,
            default_value=False,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=2077,
    serialized_end=2187,
)

_LEDGERAPIMESSAGE = _descriptor.Descriptor(
    name="LedgerApiMessage",
    full_name="fetch.aea.LedgerApi.LedgerApiMessage",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="message_id",
            full_name="fetch.aea.LedgerApi.LedgerApiMessage.message_id",
            index=0,
            number=1,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="dialogue_starter_reference",
            full_name="fetch.aea.LedgerApi.LedgerApiMessage.dialogue_starter_reference",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="dialogue_responder_reference",
            full_name="fetch.aea.LedgerApi.LedgerApiMessage.dialogue_responder_reference",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="target",
            full_name="fetch.aea.LedgerApi.LedgerApiMessage.target",
            index=3,
            number=4,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="balance",
            full_name="fetch.aea.LedgerApi.LedgerApiMessage.balance",
            index=4,
            number=5,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="error",
            full_name="fetch.aea.LedgerApi.LedgerApiMessage.error",
            index=5,
            number=6,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="get_balance",
            full_name="fetch.aea.LedgerApi.LedgerApiMessage.get_balance",
            index=6,
            number=7,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="get_raw_transaction",
            full_name="fetch.aea.LedgerApi.LedgerApiMessage.get_raw_transaction",
            index=7,
            number=8,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="get_transaction_receipt",
            full_name="fetch.aea.LedgerApi.LedgerApiMessage.get_transaction_receipt",
            index=8,
            number=9,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="raw_transaction",
            full_name="fetch.aea.LedgerApi.LedgerApiMessage.raw_transaction",
            index=9,
            number=10,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="send_signed_transaction",
            full_name="fetch.aea.LedgerApi.LedgerApiMessage.send_signed_transaction",
            index=10,
            number=11,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="transaction_digest",
            full_name="fetch.aea.LedgerApi.LedgerApiMessage.transaction_digest",
            index=11,
            number=12,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="transaction_receipt",
            full_name="fetch.aea.LedgerApi.LedgerApiMessage.transaction_receipt",
            index=12,
            number=13,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[
        _LEDGERAPIMESSAGE_RAWTRANSACTION,
        _LEDGERAPIMESSAGE_SIGNEDTRANSACTION,
        _LEDGERAPIMESSAGE_TERMS,
        _LEDGERAPIMESSAGE_TRANSACTIONDIGEST,
        _LEDGERAPIMESSAGE_TRANSACTIONRECEIPT,
        _LEDGERAPIMESSAGE_GET_BALANCE_PERFORMATIVE,
        _LEDGERAPIMESSAGE_GET_RAW_TRANSACTION_PERFORMATIVE,
        _LEDGERAPIMESSAGE_SEND_SIGNED_TRANSACTION_PERFORMATIVE,
        _LEDGERAPIMESSAGE_GET_TRANSACTION_RECEIPT_PERFORMATIVE,
        _LEDGERAPIMESSAGE_BALANCE_PERFORMATIVE,
        _LEDGERAPIMESSAGE_RAW_TRANSACTION_PERFORMATIVE,
        _LEDGERAPIMESSAGE_TRANSACTION_DIGEST_PERFORMATIVE,
        _LEDGERAPIMESSAGE_TRANSACTION_RECEIPT_PERFORMATIVE,
        _LEDGERAPIMESSAGE_ERROR_PERFORMATIVE,
    ],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[
        _descriptor.OneofDescriptor(
            name="performative",
            full_name="fetch.aea.LedgerApi.LedgerApiMessage.performative",
            index=0,
            containing_type=None,
            fields=[],
        ),
    ],
    serialized_start=42,
    serialized_end=2203,
)

_LEDGERAPIMESSAGE_RAWTRANSACTION.containing_type = _LEDGERAPIMESSAGE
_LEDGERAPIMESSAGE_SIGNEDTRANSACTION.containing_type = _LEDGERAPIMESSAGE
_LEDGERAPIMESSAGE_TERMS.containing_type = _LEDGERAPIMESSAGE
_LEDGERAPIMESSAGE_TRANSACTIONDIGEST.containing_type = _LEDGERAPIMESSAGE
_LEDGERAPIMESSAGE_TRANSACTIONRECEIPT.containing_type = _LEDGERAPIMESSAGE
_LEDGERAPIMESSAGE_GET_BALANCE_PERFORMATIVE.containing_type = _LEDGERAPIMESSAGE
_LEDGERAPIMESSAGE_GET_RAW_TRANSACTION_PERFORMATIVE.fields_by_name[
    "terms"
].message_type = _LEDGERAPIMESSAGE_TERMS
_LEDGERAPIMESSAGE_GET_RAW_TRANSACTION_PERFORMATIVE.containing_type = _LEDGERAPIMESSAGE
_LEDGERAPIMESSAGE_SEND_SIGNED_TRANSACTION_PERFORMATIVE.fields_by_name[
    "signed_transaction"
].message_type = _LEDGERAPIMESSAGE_SIGNEDTRANSACTION
_LEDGERAPIMESSAGE_SEND_SIGNED_TRANSACTION_PERFORMATIVE.containing_type = (
    _LEDGERAPIMESSAGE
)
_LEDGERAPIMESSAGE_GET_TRANSACTION_RECEIPT_PERFORMATIVE.fields_by_name[
    "transaction_digest"
].message_type = _LEDGERAPIMESSAGE_TRANSACTIONDIGEST
_LEDGERAPIMESSAGE_GET_TRANSACTION_RECEIPT_PERFORMATIVE.containing_type = (
    _LEDGERAPIMESSAGE
)
_LEDGERAPIMESSAGE_BALANCE_PERFORMATIVE.containing_type = _LEDGERAPIMESSAGE
_LEDGERAPIMESSAGE_RAW_TRANSACTION_PERFORMATIVE.fields_by_name[
    "raw_transaction"
].message_type = _LEDGERAPIMESSAGE_RAWTRANSACTION
_LEDGERAPIMESSAGE_RAW_TRANSACTION_PERFORMATIVE.containing_type = _LEDGERAPIMESSAGE
_LEDGERAPIMESSAGE_TRANSACTION_DIGEST_PERFORMATIVE.fields_by_name[
    "transaction_digest"
].message_type = _LEDGERAPIMESSAGE_TRANSACTIONDIGEST
_LEDGERAPIMESSAGE_TRANSACTION_DIGEST_PERFORMATIVE.containing_type = _LEDGERAPIMESSAGE
_LEDGERAPIMESSAGE_TRANSACTION_RECEIPT_PERFORMATIVE.fields_by_name[
    "transaction_receipt"
].message_type = _LEDGERAPIMESSAGE_TRANSACTIONRECEIPT
_LEDGERAPIMESSAGE_TRANSACTION_RECEIPT_PERFORMATIVE.containing_type = _LEDGERAPIMESSAGE
_LEDGERAPIMESSAGE_ERROR_PERFORMATIVE.containing_type = _LEDGERAPIMESSAGE
_LEDGERAPIMESSAGE.fields_by_name[
    "balance"
].message_type = _LEDGERAPIMESSAGE_BALANCE_PERFORMATIVE
_LEDGERAPIMESSAGE.fields_by_name[
    "error"
].message_type = _LEDGERAPIMESSAGE_ERROR_PERFORMATIVE
_LEDGERAPIMESSAGE.fields_by_name[
    "get_balance"
].message_type = _LEDGERAPIMESSAGE_GET_BALANCE_PERFORMATIVE
_LEDGERAPIMESSAGE.fields_by_name[
    "get_raw_transaction"
].message_type = _LEDGERAPIMESSAGE_GET_RAW_TRANSACTION_PERFORMATIVE
_LEDGERAPIMESSAGE.fields_by_name[
    "get_transaction_receipt"
].message_type = _LEDGERAPIMESSAGE_GET_TRANSACTION_RECEIPT_PERFORMATIVE
_LEDGERAPIMESSAGE.fields_by_name[
    "raw_transaction"
].message_type = _LEDGERAPIMESSAGE_RAW_TRANSACTION_PERFORMATIVE
_LEDGERAPIMESSAGE.fields_by_name[
    "send_signed_transaction"
].message_type = _LEDGERAPIMESSAGE_SEND_SIGNED_TRANSACTION_PERFORMATIVE
_LEDGERAPIMESSAGE.fields_by_name[
    "transaction_digest"
].message_type = _LEDGERAPIMESSAGE_TRANSACTION_DIGEST_PERFORMATIVE
_LEDGERAPIMESSAGE.fields_by_name[
    "transaction_receipt"
].message_type = _LEDGERAPIMESSAGE_TRANSACTION_RECEIPT_PERFORMATIVE
_LEDGERAPIMESSAGE.oneofs_by_name["performative"].fields.append(
    _LEDGERAPIMESSAGE.fields_by_name["balance"]
)
_LEDGERAPIMESSAGE.fields_by_name[
    "balance"
].containing_oneof = _LEDGERAPIMESSAGE.oneofs_by_name["performative"]
_LEDGERAPIMESSAGE.oneofs_by_name["performative"].fields.append(
    _LEDGERAPIMESSAGE.fields_by_name["error"]
)
_LEDGERAPIMESSAGE.fields_by_name[
    "error"
].containing_oneof = _LEDGERAPIMESSAGE.oneofs_by_name["performative"]
_LEDGERAPIMESSAGE.oneofs_by_name["performative"].fields.append(
    _LEDGERAPIMESSAGE.fields_by_name["get_balance"]
)
_LEDGERAPIMESSAGE.fields_by_name[
    "get_balance"
].containing_oneof = _LEDGERAPIMESSAGE.oneofs_by_name["performative"]
_LEDGERAPIMESSAGE.oneofs_by_name["performative"].fields.append(
    _LEDGERAPIMESSAGE.fields_by_name["get_raw_transaction"]
)
_LEDGERAPIMESSAGE.fields_by_name[
    "get_raw_transaction"
].containing_oneof = _LEDGERAPIMESSAGE.oneofs_by_name["performative"]
_LEDGERAPIMESSAGE.oneofs_by_name["performative"].fields.append(
    _LEDGERAPIMESSAGE.fields_by_name["get_transaction_receipt"]
)
_LEDGERAPIMESSAGE.fields_by_name[
    "get_transaction_receipt"
].containing_oneof = _LEDGERAPIMESSAGE.oneofs_by_name["performative"]
_LEDGERAPIMESSAGE.oneofs_by_name["performative"].fields.append(
    _LEDGERAPIMESSAGE.fields_by_name["raw_transaction"]
)
_LEDGERAPIMESSAGE.fields_by_name[
    "raw_transaction"
].containing_oneof = _LEDGERAPIMESSAGE.oneofs_by_name["performative"]
_LEDGERAPIMESSAGE.oneofs_by_name["performative"].fields.append(
    _LEDGERAPIMESSAGE.fields_by_name["send_signed_transaction"]
)
_LEDGERAPIMESSAGE.fields_by_name[
    "send_signed_transaction"
].containing_oneof = _LEDGERAPIMESSAGE.oneofs_by_name["performative"]
_LEDGERAPIMESSAGE.oneofs_by_name["performative"].fields.append(
    _LEDGERAPIMESSAGE.fields_by_name["transaction_digest"]
)
_LEDGERAPIMESSAGE.fields_by_name[
    "transaction_digest"
].containing_oneof = _LEDGERAPIMESSAGE.oneofs_by_name["performative"]
_LEDGERAPIMESSAGE.oneofs_by_name["performative"].fields.append(
    _LEDGERAPIMESSAGE.fields_by_name["transaction_receipt"]
)
_LEDGERAPIMESSAGE.fields_by_name[
    "transaction_receipt"
].containing_oneof = _LEDGERAPIMESSAGE.oneofs_by_name["performative"]
DESCRIPTOR.message_types_by_name["LedgerApiMessage"] = _LEDGERAPIMESSAGE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

LedgerApiMessage = _reflection.GeneratedProtocolMessageType(
    "LedgerApiMessage",
    (_message.Message,),
    {
        "RawTransaction": _reflection.GeneratedProtocolMessageType(
            "RawTransaction",
            (_message.Message,),
            {
                "DESCRIPTOR": _LEDGERAPIMESSAGE_RAWTRANSACTION,
                "__module__": "ledger_api_pb2"
                # @@protoc_insertion_point(class_scope:fetch.aea.LedgerApi.LedgerApiMessage.RawTransaction)
            },
        ),
        "SignedTransaction": _reflection.GeneratedProtocolMessageType(
            "SignedTransaction",
            (_message.Message,),
            {
                "DESCRIPTOR": _LEDGERAPIMESSAGE_SIGNEDTRANSACTION,
                "__module__": "ledger_api_pb2"
                # @@protoc_insertion_point(class_scope:fetch.aea.LedgerApi.LedgerApiMessage.SignedTransaction)
            },
        ),
        "Terms": _reflection.GeneratedProtocolMessageType(
            "Terms",
            (_message.Message,),
            {
                "DESCRIPTOR": _LEDGERAPIMESSAGE_TERMS,
                "__module__": "ledger_api_pb2"
                # @@protoc_insertion_point(class_scope:fetch.aea.LedgerApi.LedgerApiMessage.Terms)
            },
        ),
        "TransactionDigest": _reflection.GeneratedProtocolMessageType(
            "TransactionDigest",
            (_message.Message,),
            {
                "DESCRIPTOR": _LEDGERAPIMESSAGE_TRANSACTIONDIGEST,
                "__module__": "ledger_api_pb2"
                # @@protoc_insertion_point(class_scope:fetch.aea.LedgerApi.LedgerApiMessage.TransactionDigest)
            },
        ),
        "TransactionReceipt": _reflection.GeneratedProtocolMessageType(
            "TransactionReceipt",
            (_message.Message,),
            {
                "DESCRIPTOR": _LEDGERAPIMESSAGE_TRANSACTIONRECEIPT,
                "__module__": "ledger_api_pb2"
                # @@protoc_insertion_point(class_scope:fetch.aea.LedgerApi.LedgerApiMessage.TransactionReceipt)
            },
        ),
        "Get_Balance_Performative": _reflection.GeneratedProtocolMessageType(
            "Get_Balance_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _LEDGERAPIMESSAGE_GET_BALANCE_PERFORMATIVE,
                "__module__": "ledger_api_pb2"
                # @@protoc_insertion_point(class_scope:fetch.aea.LedgerApi.LedgerApiMessage.Get_Balance_Performative)
            },
        ),
        "Get_Raw_Transaction_Performative": _reflection.GeneratedProtocolMessageType(
            "Get_Raw_Transaction_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _LEDGERAPIMESSAGE_GET_RAW_TRANSACTION_PERFORMATIVE,
                "__module__": "ledger_api_pb2"
                # @@protoc_insertion_point(class_scope:fetch.aea.LedgerApi.LedgerApiMessage.Get_Raw_Transaction_Performative)
            },
        ),
        "Send_Signed_Transaction_Performative": _reflection.GeneratedProtocolMessageType(
            "Send_Signed_Transaction_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _LEDGERAPIMESSAGE_SEND_SIGNED_TRANSACTION_PERFORMATIVE,
                "__module__": "ledger_api_pb2"
                # @@protoc_insertion_point(class_scope:fetch.aea.LedgerApi.LedgerApiMessage.Send_Signed_Transaction_Performative)
            },
        ),
        "Get_Transaction_Receipt_Performative": _reflection.GeneratedProtocolMessageType(
            "Get_Transaction_Receipt_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _LEDGERAPIMESSAGE_GET_TRANSACTION_RECEIPT_PERFORMATIVE,
                "__module__": "ledger_api_pb2"
                # @@protoc_insertion_point(class_scope:fetch.aea.LedgerApi.LedgerApiMessage.Get_Transaction_Receipt_Performative)
            },
        ),
        "Balance_Performative": _reflection.GeneratedProtocolMessageType(
            "Balance_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _LEDGERAPIMESSAGE_BALANCE_PERFORMATIVE,
                "__module__": "ledger_api_pb2"
                # @@protoc_insertion_point(class_scope:fetch.aea.LedgerApi.LedgerApiMessage.Balance_Performative)
            },
        ),
        "Raw_Transaction_Performative": _reflection.GeneratedProtocolMessageType(
            "Raw_Transaction_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _LEDGERAPIMESSAGE_RAW_TRANSACTION_PERFORMATIVE,
                "__module__": "ledger_api_pb2"
                # @@protoc_insertion_point(class_scope:fetch.aea.LedgerApi.LedgerApiMessage.Raw_Transaction_Performative)
            },
        ),
        "Transaction_Digest_Performative": _reflection.GeneratedProtocolMessageType(
            "Transaction_Digest_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _LEDGERAPIMESSAGE_TRANSACTION_DIGEST_PERFORMATIVE,
                "__module__": "ledger_api_pb2"
                # @@protoc_insertion_point(class_scope:fetch.aea.LedgerApi.LedgerApiMessage.Transaction_Digest_Performative)
            },
        ),
        "Transaction_Receipt_Performative": _reflection.GeneratedProtocolMessageType(
            "Transaction_Receipt_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _LEDGERAPIMESSAGE_TRANSACTION_RECEIPT_PERFORMATIVE,
                "__module__": "ledger_api_pb2"
                # @@protoc_insertion_point(class_scope:fetch.aea.LedgerApi.LedgerApiMessage.Transaction_Receipt_Performative)
            },
        ),
        "Error_Performative": _reflection.GeneratedProtocolMessageType(
            "Error_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _LEDGERAPIMESSAGE_ERROR_PERFORMATIVE,
                "__module__": "ledger_api_pb2"
                # @@protoc_insertion_point(class_scope:fetch.aea.LedgerApi.LedgerApiMessage.Error_Performative)
            },
        ),
        "DESCRIPTOR": _LEDGERAPIMESSAGE,
        "__module__": "ledger_api_pb2"
        # @@protoc_insertion_point(class_scope:fetch.aea.LedgerApi.LedgerApiMessage)
    },
)
_sym_db.RegisterMessage(LedgerApiMessage)
_sym_db.RegisterMessage(LedgerApiMessage.RawTransaction)
_sym_db.RegisterMessage(LedgerApiMessage.SignedTransaction)
_sym_db.RegisterMessage(LedgerApiMessage.Terms)
_sym_db.RegisterMessage(LedgerApiMessage.TransactionDigest)
_sym_db.RegisterMessage(LedgerApiMessage.TransactionReceipt)
_sym_db.RegisterMessage(LedgerApiMessage.Get_Balance_Performative)
_sym_db.RegisterMessage(LedgerApiMessage.Get_Raw_Transaction_Performative)
_sym_db.RegisterMessage(LedgerApiMessage.Send_Signed_Transaction_Performative)
_sym_db.RegisterMessage(LedgerApiMessage.Get_Transaction_Receipt_Performative)
_sym_db.RegisterMessage(LedgerApiMessage.Balance_Performative)
_sym_db.RegisterMessage(LedgerApiMessage.Raw_Transaction_Performative)
_sym_db.RegisterMessage(LedgerApiMessage.Transaction_Digest_Performative)
_sym_db.RegisterMessage(LedgerApiMessage.Transaction_Receipt_Performative)
_sym_db.RegisterMessage(LedgerApiMessage.Error_Performative)


# @@protoc_insertion_point(module_scope)
