pre {
    background-color: #f8f8f7;
}

code {
    background-color: #0083fb;
}

/* this doesn't work now
.md-nav__link {
  text-transform: uppercase;
  color: #0083fb;
}
*/

/* Katharine's css additions */
.md-header,
.md-tabs,
.md-footer-meta,
.md-footer-nav,
.md-footer-nav__inner {
    background-color: #172b6e;
}

.md-nav__title {
    color: #172b6e;
}

.md-icon {
    ./assets/images/favicon.ico;
}

/* Needed so that Mermaid UML diagrams don't end up being massively tall */
svg{
    height: auto;

}