hr {
    border: 1px solid #FFFFFF;
    opacity: 1;
}

nav {
    background: #1E2844 0% 0% no-repeat padding-box;
    height: 80px;
}

body {
    background: #161E33 0% 0% no-repeat padding-box;
}

table {
    color: #A1B6C0 !important;
    border-color: #A1B6C0;
}

thead {
    color: #FFFFFF;
}

button {
    height: 26px;
    border-radius: 2px;
    text-align: center;
    letter-spacing: -0.14px;
    color: #FFFFFF;
    border: none;
}

button:hover {
    color: #161E33;
    background: #FFFFFF;
}

input[type="text"] {
    height: 23px;
    border-radius: 2px;
    border: 1px solid #FFFFFF;
    color: #FFFFFF;
    background: transparent;
    text-indent: 4px;
}

input[type="radio"] {
    width: 19px;
    height: 19px;
    border: 1px solid #FFFFFF;
}

input:focus {
    outline: none !important;
}

::selection {
    color: #FFFFFF;
    background: #b3d7ff;
}

/* header items */
.fetchai-logo {
    width: auto;
    margin-left: 40px;
    top: 20px;
    left: 47px;
    width: 164px;
    height: 51px;
}

.app-title {
    text-align: right;
    letter-spacing: -0.2px;
    color: #FFFFFF !important;
}

/* boxes */
.app-box {
    border-radius: 4px;
}

.app-box-green {
    background: rgba(0, 185, 162, 0.1) 0% 0% no-repeat padding-box;
}

.app-box-blue {
    background: rgba(76, 130, 220, 0.1) 0% 0% no-repeat padding-box;
}

.app-box-violet {
    background: rgba(164, 90, 149, 0.1) 0% 0% no-repeat padding-box;
}

.box-title {
    text-align: center;
    letter-spacing: -0.18px;
}

/* buttons */
.app-btn-green {
    background-color: rgba(0, 185, 162, 1);
}

.app-btn-blue {
    background-color: rgba(76, 130, 220, 1);
}

.app-btn-s {
    width: 82px;
}

.app-btn-m {
    width: 110px;
}

.app-btn-l {
    width: 172px;
}

/* inputs */
.app-create-agent-input {
    width: 183px;
}

.app-search-input {
    width: 203px;
}

.app-connection-id-input {
    width: 183px;
}

/* miscellaneous */
.app-output {
    height: 230px;
    border: 1px solid #FFFFFF;
    border-radius: 4px;
    overflow: auto !important;
}

.error {
    position: fixed;
    /* Sit on top of the page content */
    width: 100%;
    /* Full width (cover the whole page) */
    height: 10%;
    /* A bit of the page */
    bottom: 0;
    visibility: hidden;
    border: 1px solid lightgrey;
    border-radius: 0px;
    background-color: #fbbd;
    text-align: center;
    font-weight: bold;
    opacity: 50%;
}
